import java.awt.Cursor;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.PrintWriter;
import java.util.Scanner;

import javax.swing.*;
// import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * @author Michael A. Covington 2013
 * Released to the public domain without warranty
 * 
 * Word frequency table program with a graphical user interface.
 * Intended as a sample of how a typical piece of scientific
 * software might be implemented in Java Swing.
 *
 */
public class CovWordFreqWin extends JFrame {

	private static final long serialVersionUID = -3639894089629317080L;

	public static void main(String[] args) { 
		
		// Use the native look and feel if possible instead of Java's own.
		// This is optional.
		try 
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} 
		catch (Exception e) 
		{
			// do nothing, take the default
		}
		
		// Launch it...
		new CovWordFreqWin(); 
	}
	
	private JTextArea textResults;
	private JScrollPane scroll;
	private JMenuBar menuBar;
	private JMenu menuFile;
	private JMenuItem menuFileOpen;
	private JMenuItem menuFileSave;
	private JMenuItem menuFileExit;
	private JMenu menuHelp;
	private JMenuItem menuHelpAbout;
	
	public CovWordFreqWin() // main window constructor
	{
		// Main window
		this.setSize(600,400);
		this.setLocationRelativeTo(null);  // center on screen
		this.setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		this.setTitle("Covington Innovations Word Frequency Program");
		
		// Menu bar
		menuBar = new JMenuBar();

		// File menu
		menuFile = new JMenu("File");
		menuFile.setMnemonic(KeyEvent.VK_F);  // Alt-F
		
		menuFileOpen = new JMenuItem("Open and analyze...");
		menuFileOpen.setMnemonic(KeyEvent.VK_O);
		menuFileOpen.addActionListener(new MainWindowEventDispatcher());
		menuFile.add(menuFileOpen);

		menuFileSave = new JMenuItem("Save to file...");
		menuFileSave.setMnemonic(KeyEvent.VK_S);
		menuFileSave.addActionListener(new MainWindowEventDispatcher());
		menuFileSave.setEnabled(false);
		menuFile.add(menuFileSave);
		
		menuFileExit = new JMenuItem("Exit");
		menuFileExit.setMnemonic(KeyEvent.VK_X);
		menuFileExit.addActionListener(new MainWindowEventDispatcher());
		menuFile.add(menuFileExit);
		
		menuBar.add(menuFile);
		
		// Help menu
		menuHelp = new JMenu("Help");
		menuHelp.setMnemonic(KeyEvent.VK_H);  // Alt-H
		
		menuHelpAbout = new JMenuItem("About this program...");
		menuHelpAbout.setMnemonic(KeyEvent.VK_A);
		menuHelpAbout.addActionListener(new MainWindowEventDispatcher());
		menuHelp.add(menuHelpAbout);
		
		menuBar.add(menuHelp);

		// Put menu bar in place
		
		this.setJMenuBar(menuBar);
		
		// Read-only text area that will hold output
		textResults = new JTextArea(10,25);
		textResults.setFont(new Font("Monospaced", Font.BOLD, 14));
		textResults.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		textResults.setLineWrap(true);
		textResults.setEditable(false);
		
		// Put the text area on the main window as scrollable
		scroll = new JScrollPane(
				textResults,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		this.add(scroll);
		
		// Show the main window
		this.setVisible(true);
	}
	
	// Event handler for all components on main window
	private class MainWindowEventDispatcher implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			Object source = e.getSource();
			if (source == menuFileOpen)
			{
				OpenAndProcessFile();
			}
			else if (source == menuFileSave)
			{
				SaveOutputToFile();
			}
			else if (source == menuFileExit)
			{
				System.exit(0);
			}
			else if (source == menuHelpAbout)
			{
				JOptionPane.showMessageDialog(null, 
						"This program makes and displays a word frequency table\n" +
				        "for a text file, which is assumed to be ASCII or UTF-8.\n" +
					    "The table can be saved to another text file. \n\n" +
				        "This is intended as a programming demonstration, not a\n" +
					    "finished tool for serious work.\n\n" +
					    "Michael A. Covington\nwww.covingtoninnovations.com\n");
			}
		}
	}
	
	
	void OpenAndProcessFile()
	{
		// There is no output on the screen
		textResults.setText("");
		menuFileSave.setEnabled(false);
		
		// Show an open file dialog
		textResults.append("Choose file...\n");
		final JFileChooser fc = new JFileChooser();
		// fc.setFileFilter(new FileNameExtensionFilter("Text files", "txt"));  // Not suitable outside Windows
		int returnVal = fc.showOpenDialog(this);
		
		// Bail out if cancelled
		if (returnVal != JFileChooser.APPROVE_OPTION) {
			textResults.setText("");
			return; 
		}
		
		String fname = fc.getSelectedFile().getAbsolutePath();
		this.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

		// Do the computation here
		
		// This is a low-tech approach, doing the computation on the main thread rather than in the background.
		// That is acceptable because the computation rarely takes more than a few seconds, and there is
		// no "cancel" button or anything else that might need to be responsive during the computation.
		// A more sophisticated approach would put the computation in the background using a SwingWorker.
		// See: http://docs.oracle.com/javase/6/docs/api/javax/swing/SwingWorker.html
		
		textResults.append("Processing: " + fname + "\n");
		textResults.update(textResults.getGraphics());

		// Read and process the file
		textResults.setText("");
		Scanner sc = null;
		try
		{
			sc = new Scanner(new File(fname), "UTF-8");
			
			// Make the freq table
			FreqTable ft = new FreqTable();
			while (sc.hasNextLine())
			{
				String s = sc.nextLine();
				for (String ss: Tokenizer.Tokenize(s))
				{
					ft.add(ss);
				}
			}
			
			// Display it
			for (String ss: ft.keySet())
			{
				textResults.append(String.format("%8d\t%s\n", ft.get(ss), ss));
			}
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null,  e.getMessage());
		}
		finally
		{
			if (sc != null) { sc.close(); }
		}
		
		// Return cursor to normal and allow saving
		this.setCursor(Cursor.getDefaultCursor());
		if (textResults.getText().length() == 0)
		{
			JOptionPane.showMessageDialog(null, "Nothing found in file.  Not UTF-8, maybe?");
		}
		else
		{ 
			menuFileSave.setEnabled(true); 
		}
	}
	
	void SaveOutputToFile()
	{
		final JFileChooser fc = new JFileChooser();
		// fc.setFileFilter(new FileNameExtensionFilter("Text files", "txt"));  // Not suitable outside Windows
		int returnVal = fc.showSaveDialog(this);
		
		// Bail out if cancelled
		if (returnVal != JFileChooser.APPROVE_OPTION) {
			textResults.setText("");
			return; 
		}

		// This is a quick operation and does not need to be backgrounded.
		PrintWriter out = null;
		try
		{
			out = new PrintWriter(fc.getSelectedFile());
			out.println(textResults.getText());
		}
		catch (Exception e)
		{
			JOptionPane.showMessageDialog(null,  e.getMessage());
		}
		finally
		{
			if (out != null) { out.close(); }
		}
	}
	
}
